#!/usr/bin/env python

"""
A simple (and sane) text indexing library.

Copyright (C) 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from os.path import commonprefix # to find common string prefixes

# Foundation classes.

class File:

    "A basic file abstraction."

    def __init__(self, f):
        self.f = f
        self.reset()

    def reset(self):
        pass

    def close(self):
        self.f.close()

class FileWriter(File):

    "Writing basic data types to files."

    def write_number(self, number):

        "Write 'number' to the file using a variable length encoding."

        # Negative numbers are not supported.

        if number < 0:
            raise ValueError, "Number %r is negative." % number

        # Special case: one byte containing zero.

        elif number == 0:
            self.f.write(chr(1) + chr(0))
            return

        # Write the number from least to most significant digits.

        nbytes = 0
        bytes = []

        while number != 0:
            lsd = number & 255
            bytes.append(chr(lsd))
            number = number >> 8
            nbytes += 1

        # Too large numbers are not supported.

        if nbytes > 255:
            raise ValueError, "Number %r is too large." % number

        bytes.insert(0, chr(nbytes))
        record = "".join(bytes)
        self.f.write(record)

    def write_unsigned_byte(self, number):

        "Write 'number' to the file using a single byte."

        if not (0 <= number <= 255):
            raise ValueError, "Number %r is out of range." % number

        self.f.write(chr(number))

    def write_string(self, s):

        "Write 's' to the file, recording its length."

        length = len(s)

        if not (0 <= length <= 255):
            raise ValueError, "String %r is too long." % s

        self.write_unsigned_byte(length)
        self.f.write(s)

class FileReader(File):

    "Reading basic data types from files."

    def read_number(self):

        "Read a number from the file."

        nbytes = ord(self.f.read(1))

        # Read each byte, adding it to the number.

        bytes = self.f.read(nbytes)

        i = 0
        shift = 0
        number = 0

        while i < nbytes:
            csd = ord(bytes[i])
            number += (csd << shift)
            shift += 8
            i += 1

        return number

    def read_unsigned_byte(self):

        "Read a number from the file, consuming a single byte."

        return ord(self.f.read(1))

    def read_string(self):

        "Read a string from the file."

        length = self.read_unsigned_byte()
        return self.f.read(length)

# Specific classes.

class PositionWriter(FileWriter):

    "Writing position information to files."

    def reset(self):
        self.last_docnum = 0

    def write_positions(self, docnum, positions):

        "Write for the document 'docnum' the given 'positions'."

        if docnum < self.last_docnum:
            raise ValueError, "Document number %r is less than previous number %r." % (docnum, self.last_docnum)

        # Write the document number delta.

        self.write_number(docnum - self.last_docnum)

        # Write the number of positions.

        self.write_number(len(positions))

        # Write the position deltas.

        last = 0
        for position in positions:
            pos = position - last
            self.write_number(pos)
            last = position

        self.last_docnum = docnum

    def write_all_positions(self, doc_positions):

        """
        Write all 'doc_positions' - a collection of tuples of the form (document
        number, position list) - to the file, returning the offset at which they
        were stored.
        """

        # Reset the writer and record the current file offset.

        self.reset()
        offset = self.f.tell()

        # Write the number of documents.

        self.write_number(len(doc_positions))

        # Write the positions.

        for docnum, positions in doc_positions:
            self.write_positions(docnum, positions)

        return offset

class PositionReader(FileReader):

    "Reading position information from files."

    def reset(self):
        self.last_docnum = 0

    def read_positions(self):

        "Read positions, returning a document number and a list of positions."

        # Read the document number delta and add it to the last number.

        self.last_docnum += self.read_number()

        # Read the number of positions.

        npositions = self.read_number()

        # Read the position deltas, adding each previous position to get the
        # appropriate collection of absolute positions.

        i = 0
        last = 0
        positions = []

        while i < npositions:
            last += self.read_number()
            positions.append(last)
            i += 1

        return self.last_docnum, positions

    def read_all_positions(self, offset):

        """
        Read all positions from 'offset', seeking to that position in the file
        before reading.
        """

        self.reset()
        self.f.seek(offset)

        # Read the number of documents.

        ndocuments = self.read_number()

        # Read all records.

        i = 0
        doc_positions = []

        while i < ndocuments:
            doc_positions.append(self.read_positions())
            i += 1

        return doc_positions

class TermWriter(FileWriter):

    "Writing term information to files."

    def reset(self):
        self.last_term = ""
        self.last_offset = 0

    def write_term(self, term, offset):

        """
        Write the given 'term' and its position file 'offset' to the term
        information file.
        """

        # Too long terms are not currently supported.

        if len(term) > 255:
            raise ValueError, "Term %r is too long." % term

        # Write the prefix length and term suffix.

        common = len(commonprefix([self.last_term, term]))
        suffix = term[common:]

        self.write_unsigned_byte(common)
        self.write_string(suffix)

        # Write the offset delta.

        self.write_number(offset - self.last_offset)

        self.last_term = term
        self.last_offset = offset

class TermReader(FileReader):

    "Reading term information from files."

    def reset(self):
        self.last_term = ""
        self.last_offset = 0

    def read_term(self):

        """
        Read a term and its position file offset from the term information file.
        """

        # Read the prefix length and term suffix.

        common = self.read_unsigned_byte()
        suffix = self.read_string()

        self.last_term = self.last_term[:common] + suffix

        # Read the offset delta.

        self.last_offset += self.read_number()

        return self.last_term, self.last_offset

# vim: tabstop=4 expandtab shiftwidth=4
